/*
 * This file is part of XFellow.
 *
 * gui.c - general functions for the GUI.
 * Copyright 2000-2001 by Dan Sutherland <dan@dillinja.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include <gtk/gtk.h>
#include <stdlib.h>

#include "defs.h"
#include "portable.h"
#include "sound.h"
#include "cpu.h"
#include "listtree.h"
#include "gameport.h"
#include "config.h"
#include "fellow.h"

#include "interface.h"
#include "gui.h"

/* the fellow config instance used by the gui */
cfg *xcfg;

/* instances of all gui windows */
GtkWidget *main_win;
GtkWidget *conf_win;

BOOLE quit_emu;
FselInfo fsel_info;

/* name of currently loaded config file */
STR current_cfg_filename[MAX_FILENAME_LEN];

/* some constants for converting mem sizes to/from strings */
#define MEM_SIZES_COUNT 9

STR memTxt[MEM_SIZES_COUNT][8] = {"0 KB", "256 KB", "512 KB", "768 KB", "1024 KB", "1280 KB", "1536 KB",
			"1792 KB", "2048 KB"};
ULO memSizes[MEM_SIZES_COUNT] = {0, 0x40000, 0x80000, 0xc0000, 0x100000, 0x140000, 0x180000,
			0x1c0000, 0x200000};

#define GAMEPORT_SETTINGS_COUNT 6
STR gpTxt[GAMEPORT_SETTINGS_COUNT][18] = {"None", "Keyboard Layout 1", "Keyboard Layout 2", "Joystick 1",
			"Joystick 2", "Mouse"};

/* The emulator core sometimes calls this function while it is running in order to
 * get some important information to the user - however GTK+ is disabled while the
 * emulation is running so we cannot display a dialogue box. This needs addressing,
 * for the moment it just outputs the message to stdout.
 */
void wguiRequester(STR *line1, STR *line2, STR *line3)
{
	printf("wguiRequester:\n%s\n%s\n%s\n", line1, line2, line3);
}

void wguiStartup(void)
{
	fellowAddLog("wguiStartup(): initialising\n");

	strcpy(current_cfg_filename, "");

	xcfg = cfgManagerGetCurrentConfig(&cfg_manager);
	gtk_init(NULL, NULL); /* TODO: parse command line */
	if((main_win = create_winMain()) == NULL) {
		fellowAddLog("wguiStartup: FATAL: failed to create main window\n");
		exit(1); // FIXME: nicer way to exit?
	}

	if((conf_win = create_winConf()) == NULL) {
		fellowAddLog("wguiStartup: FATAL: failed to create conf window\n");
		exit(1);
	}
}

void wguiShutdown(void)
{
	/* nothing to do here, since clicking the quit button destroys the GUI */
}

gint emuStart()
{
	disk2cfg();
	fellowPreStartReset(fellowGetPreStartReset() |
			cfgManagerConfigurationActivate(&cfg_manager));
	gtk_widget_hide(main_win);
	gtk_main_quit();
	fellowAddLog("gui: emuStart(): starting emulator\n");
	fellowEmulationStart();
	fellowRun();
	fellowEmulationStop();
}

BOOLE wguiEnter(void)
{
	fellowAddLog("wguiEnter(): entering\n");
	fprintf(stderr, "wguiEnter(): entering\n");
	quit_emu = FALSE;

	cfg2disk();
	gtk_widget_show(main_win);
	gtk_main();

	fellowAddLog("wguiEnter(): leaving\n");

	return quit_emu;
}

/* show a file selector, and if a file is selected, put its path in txtbox */
void guiGetFileName(GtkWidget *txtbox)
{
	static GtkWidget *fsel;

	fsel = gtk_file_selection_new("Select File...");

	fsel_info.fsel = fsel;
	fsel_info.txtbox = txtbox;
	fsel_info.mode = FSEL_MODE_FILE_TO_ENTRY;

	gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(fsel)->ok_button), "clicked",
			GTK_SIGNAL_FUNC(onFselOKClicked), &fsel_info);
	gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(fsel)->cancel_button), "clicked",
			GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(fsel));

	gtk_widget_show(fsel);
}

void onFselOKClicked(GtkWidget *w, FselInfo *inf)
{
	STR *file;

	if((file = malloc(MAX_FILENAME_LEN)) == NULL) {
		fellowAddLog("onFselOKClicked(): FATAL: couldn't allocate temporary string\n");
		exit(1);
	}

	file = gtk_file_selection_get_filename(GTK_FILE_SELECTION(inf->fsel));

	switch(inf->mode) {
	case FSEL_MODE_FILE_TO_ENTRY:
		gtk_entry_set_text(GTK_ENTRY(inf->txtbox), file);
		break;
	case FSEL_MODE_LOAD_CFG:
		cfgLoadFromFilename(xcfg, file);
		cfg2disk();
		strcpy(current_cfg_filename, file);
		break;
	case FSEL_MODE_SAVE_CFG:
		disk2cfg();
		cfgSaveToFilename(xcfg, file);
		break;
	}

	gtk_widget_destroy(inf->fsel);
}

void guiLoadSaveCfg(BOOLE load)
{
	static GtkWidget *fsel;
	STR *tmpstr;

	if((tmpstr = malloc(MAX_FILENAME_LEN)) == NULL) {
		fellowAddLog("onFselOKClicked(): FATAL: couldn't allocate temporary string\n");
		exit(1);
	}

	sprintf(tmpstr, "%s config file...", load ? "load" : "save");
	fsel = gtk_file_selection_new(tmpstr);

	fsel_info.fsel = fsel;
	fsel_info.txtbox = NULL;
	fsel_info.mode = load ? FSEL_MODE_LOAD_CFG : FSEL_MODE_SAVE_CFG;

	gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(fsel)->ok_button), "clicked",
			GTK_SIGNAL_FUNC(onFselOKClicked), &fsel_info);
	gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(fsel)->cancel_button), "clicked",
			GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(fsel));

	gtk_widget_show(fsel);
}

/* Apply disk image setting to current config */
void disk2cfg()
{
	STR *tmpstr;

	tmpstr = malloc(MAX_FILENAME_LEN);
	if(tmpstr == NULL) {
		fellowAddLog("disk2cfg(): FATAL: couldn't allocate temporary buffer\n");
		exit(1);
	}
	tmpstr = gtk_entry_get_text(GTK_ENTRY(txtDF0));
	cfgSetDiskImage(xcfg, 0, tmpstr);
	tmpstr = gtk_entry_get_text(GTK_ENTRY(txtDF1));
	cfgSetDiskImage(xcfg, 1, tmpstr);
	tmpstr = gtk_entry_get_text(GTK_ENTRY(txtDF2));
	cfgSetDiskImage(xcfg, 2, tmpstr);
	tmpstr = gtk_entry_get_text(GTK_ENTRY(txtDF3));
	cfgSetDiskImage(xcfg, 3, tmpstr);

	free(tmpstr);
}

/* Output disk image settings from current config to main window */
void cfg2disk()
{
	STR *tmpstr;

	tmpstr = malloc(MAX_FILENAME_LEN);
	if(tmpstr == NULL) {
		fellowAddLog("disk2cfg(): FATAL: couldn't allocate temporary buffer\n");
		exit(1);
	}

	tmpstr = cfgGetDiskImage(xcfg, 0);
	gtk_entry_set_text(GTK_ENTRY(txtDF0), tmpstr);
	tmpstr = cfgGetDiskImage(xcfg, 1);
	gtk_entry_set_text(GTK_ENTRY(txtDF1), tmpstr);
	tmpstr = cfgGetDiskImage(xcfg, 2);
	gtk_entry_set_text(GTK_ENTRY(txtDF2), tmpstr);
	tmpstr = cfgGetDiskImage(xcfg, 3);
	gtk_entry_set_text(GTK_ENTRY(txtDF3), tmpstr);

	/* GTK+ takes care of our tmpstr value after we pass it to gtk_entry_set_text,
		so there is no need to free it (in fact it will crash if we do) */
}

/* Get all current settings from the GUI and apply to the current config
 */
void gui2cfg()
{
	STR *tmpstr;
	int i;
	BOOLE ok;

	tmpstr = malloc(MAX_FILENAME_LEN);
	if(tmpstr == NULL) {
		fellowAddLog("gui2cfg(): FATAL: couldn't allocate temporary string buffer\n");
		exit(1);
	}

	/* ----- CPU PAGE ----- */

	if(GTK_TOGGLE_BUTTON(radCPU68000)->active) {
		cfgSetCPUType(xcfg, M68000);
	} else if(GTK_TOGGLE_BUTTON(radCPU68010)->active) {
		cfgSetCPUType(xcfg, M68010);
	} else if(GTK_TOGGLE_BUTTON(radCPU68020)->active) {
		cfgSetCPUType(xcfg, M68020);
	} else if(GTK_TOGGLE_BUTTON(radCPU68030)->active) {
		cfgSetCPUType(xcfg, M68030);
	} else if(GTK_TOGGLE_BUTTON(radCPU68EC20)->active) {
		cfgSetCPUType(xcfg, M68EC20);
	} else if(GTK_TOGGLE_BUTTON(radCPU68EC30)->active) {
		cfgSetCPUType(xcfg, M68EC30);
	}

	if(GTK_TOGGLE_BUTTON(radSpeed35)->active) {
		cfgSetCPUSpeed(xcfg, 1);
	} else if(GTK_TOGGLE_BUTTON(radSpeed7)->active) {
		cfgSetCPUSpeed(xcfg, 2);
	} else if(GTK_TOGGLE_BUTTON(radSpeed14)->active) {
		cfgSetCPUSpeed(xcfg, 4);
	} else if(GTK_TOGGLE_BUTTON(radSpeed28)->active) {
		cfgSetCPUSpeed(xcfg, 8);
	}

	/* ----- FLOPPY PAGE ----- */

	tmpstr = gtk_entry_get_text(GTK_ENTRY(txtConfDF0));
	cfgSetDiskImage(xcfg, 0, tmpstr);
	tmpstr = gtk_entry_get_text(GTK_ENTRY(txtConfDF1));
	cfgSetDiskImage(xcfg, 1, tmpstr);
	tmpstr = gtk_entry_get_text(GTK_ENTRY(txtConfDF2));
	cfgSetDiskImage(xcfg, 2, tmpstr);
	tmpstr = gtk_entry_get_text(GTK_ENTRY(txtConfDF3));
	cfgSetDiskImage(xcfg, 3, tmpstr);

	cfgSetDiskEnabled(xcfg, 0, GTK_TOGGLE_BUTTON(chkImgEnabled0)->active);
	cfgSetDiskEnabled(xcfg, 1, GTK_TOGGLE_BUTTON(chkImgEnabled1)->active);
	cfgSetDiskEnabled(xcfg, 2, GTK_TOGGLE_BUTTON(chkImgEnabled2)->active);
	cfgSetDiskEnabled(xcfg, 3, GTK_TOGGLE_BUTTON(chkImgEnabled3)->active);

	cfgSetDiskReadOnly(xcfg, 0, GTK_TOGGLE_BUTTON(chkImgRO0)->active);
	cfgSetDiskReadOnly(xcfg, 1, GTK_TOGGLE_BUTTON(chkImgRO1)->active);
	cfgSetDiskReadOnly(xcfg, 2, GTK_TOGGLE_BUTTON(chkImgRO2)->active);
	cfgSetDiskReadOnly(xcfg, 3, GTK_TOGGLE_BUTTON(chkImgRO3)->active);

	cfgSetDiskFast(xcfg, GTK_TOGGLE_BUTTON(chkFastDiskDMA)->active);

	/* ----- MEMORY PAGE ----- */

	tmpstr = gtk_entry_get_text(GTK_ENTRY(txtMemChipEntry));
	ok = FALSE;
	for(i = 1; i <= 9; i++) {
		if(strcmp(tmpstr, memTxt[i]) == 0) {
			cfgSetChipSize(xcfg, memSizes[i]);
			ok = TRUE;
		}
	}

	/* paranoid check */
	if(!ok) {
		fellowAddLog("gui2cfg(): impossible value for txtMemChipEntry - defaulting to 512KB\n");
		cfgSetChipSize(xcfg, 0x80000);
	}

	tmpstr = gtk_entry_get_text(GTK_ENTRY(txtMemFastEntry));
	cfgSetFastSize(xcfg, ((tmpstr[0] - '0') * 0x100000));

	tmpstr = gtk_entry_get_text(GTK_ENTRY(txtMemBogoEntry));
	ok = FALSE;
	for(i = 0; i<= 8; i++) {
		if(strcmp(tmpstr, memTxt[i]) == 0) {
			cfgSetBogoSize(xcfg, memSizes[i]);
			ok = TRUE;
		}
	}

	/* another paranoid check */
	if(!ok) {
		fellowAddLog("gui2cfg(): impossible value for txtMemBogoEntry - defaulting to 0\n");
		cfgSetBogoSize(xcfg, 0);
	}

	/* image and key */
	tmpstr = gtk_entry_get_text(GTK_ENTRY(txtKickImage));
	cfgSetKickImage(xcfg, tmpstr);

	tmpstr = gtk_entry_get_text(GTK_ENTRY(txtKickKey));
	cfgSetKey(xcfg, tmpstr);


	/* ----- SOUND PAGE ----- */

	/* emulation type */
	if(GTK_TOGGLE_BUTTON(radSndEmuNo)->active) {
		cfgSetSoundEmulation(xcfg, SOUND_NONE);
	} else if(GTK_TOGGLE_BUTTON(radSndEmuPlay)->active) {
		cfgSetSoundEmulation(xcfg, SOUND_PLAY);
	} else {
		cfgSetSoundEmulation(xcfg, SOUND_EMULATE);
	}

	/* filter */
	if(GTK_TOGGLE_BUTTON(radFilterOrig)->active) {
		cfgSetSoundFilter(xcfg, SOUND_FILTER_ORIGINAL);
	} else if(GTK_TOGGLE_BUTTON(radFilterAlways)->active) {
		cfgSetSoundFilter(xcfg, SOUND_FILTER_ALWAYS);
	} else {
		cfgSetSoundFilter(xcfg, SOUND_FILTER_NEVER);
	}

	/* rate */
	if(GTK_TOGGLE_BUTTON(radSamRate44100)->active) {
		cfgSetSoundRate(xcfg, SOUND_44100);
	} else if(GTK_TOGGLE_BUTTON(radSamRate31300)->active) {
		cfgSetSoundRate(xcfg, SOUND_31300);
	} else if(GTK_TOGGLE_BUTTON(radSamRate22050)->active) {
		cfgSetSoundRate(xcfg, SOUND_22050);
	} else {
		cfgSetSoundRate(xcfg, SOUND_15650);
	}

	/* channels, bits, stream */
	cfgSetSoundStereo(xcfg, GTK_TOGGLE_BUTTON(radChanStereo)->active);
	cfgSetSound16Bits(xcfg, GTK_TOGGLE_BUTTON(radBits16)->active);
	cfgSetSoundWAVDump(xcfg, GTK_TOGGLE_BUTTON(chkSoundStreamWAV)->active);

	/* ----- SCREEN PAGE ----- */
	// TODO: resolution, frameskip

	/* vertical scale */
	if(GTK_TOGGLE_BUTTON(radVScaleScan)->active) {
		cfgSetVerticalScale(xcfg, 1);
		cfgSetScanlines(xcfg, TRUE);
	} else {
		cfgSetScanlines(xcfg, FALSE);
		cfgSetVerticalScale(xcfg, (GTK_TOGGLE_BUTTON(radVScaleSingle)->active) ? 1 : 2);
	}
	cfgSetDeinterlace(xcfg, GTK_TOGGLE_BUTTON(chkVScaleIntComp)->active);

	/* horizontal scale */
	cfgSetHorisontalScale(xcfg, GTK_TOGGLE_BUTTON(radHScaleSingle)->active);

	/* blitter */
	cfgSetBlitterFast(xcfg, GTK_TOGGLE_BUTTON(chkBlitImmediate)->active);
	cfgSetECSBlitter(xcfg, GTK_TOGGLE_BUTTON(chkBlitECS)->active);

	/* ----- FILESYSTEMS PAGE ----- */

	/* ----- HARDFILE PAGE ----- */

	/* ----- GAMEPORT PAGE ----- */
	tmpstr = gtk_entry_get_text(GTK_ENTRY(txtGameport1Entry));
	for(i = 0; i < GAMEPORT_SETTINGS_COUNT; i++) {
		if(strcmp(tmpstr, gpTxt[i]) == 0) {
			cfgSetGameport(xcfg, 0, i);
		}
	}

	tmpstr = gtk_entry_get_text(GTK_ENTRY(txtGameport2Entry));
	for(i = 0; i < GAMEPORT_SETTINGS_COUNT; i++) {
		if(strcmp(tmpstr, gpTxt[i]) == 0) {
			cfgSetGameport(xcfg, 1, i);
		}
	}

	/* ----- MISC PAGE ----- */

	if(GTK_TOGGLE_BUTTON(chkGenShowFPS)->active) {
		cfgSetScreenDrawLEDs(xcfg, TRUE);
	} else {
		cfgSetScreenDrawLEDs(xcfg, FALSE);
	}

	free(tmpstr);
}

void cfg2gui()
{
	STR *tmpstr;
	BOOLE ok;
	int i;

	tmpstr = malloc(MAX_FILENAME_LEN);
	if(tmpstr == NULL) {
		fellowAddLog("cfg2gui(): FATAL: couldn't allocate temporary buffer\n");
		exit(1);
	}

	/* ----- CPU PAGE ----- */
	switch(cfgGetCPUType(xcfg)) {
	case M68000:
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(radCPU68000), TRUE);
		break;
	case M68010:
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(radCPU68010), TRUE);
		break;
	case M68020:
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(radCPU68020), TRUE);
		break;
	case M68030:
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(radCPU68030), TRUE);
		break;
	case M68EC20:
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(radCPU68EC20), TRUE);
		break;
	case M68EC30:
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(radCPU68EC30), TRUE);
	}

	switch(cfgGetCPUSpeed(xcfg)) {
	case 1:
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(radSpeed35), TRUE);
		break;
	case 2:
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(radSpeed7), TRUE);
		break;
	case 4:
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(radSpeed14), TRUE);
		break;
	case 8:
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(radSpeed28), TRUE);
	}

	/* ----- FLOPPY PAGE ----- */
	tmpstr = cfgGetDiskImage(xcfg, 0);
	gtk_entry_set_text(GTK_ENTRY(txtConfDF0), tmpstr);
	tmpstr = cfgGetDiskImage(xcfg, 1);
	gtk_entry_set_text(GTK_ENTRY(txtConfDF1), tmpstr);
	tmpstr = cfgGetDiskImage(xcfg, 2);
	gtk_entry_set_text(GTK_ENTRY(txtConfDF2), tmpstr);
	tmpstr = cfgGetDiskImage(xcfg, 3);
	gtk_entry_set_text(GTK_ENTRY(txtConfDF3), tmpstr);

	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(chkImgEnabled0), cfgGetDiskEnabled(xcfg, 0));
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(chkImgEnabled1), cfgGetDiskEnabled(xcfg, 1));
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(chkImgEnabled2), cfgGetDiskEnabled(xcfg, 2));
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(chkImgEnabled3), cfgGetDiskEnabled(xcfg, 3));

	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(chkImgRO0), cfgGetDiskReadOnly(xcfg, 0));
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(chkImgRO1), cfgGetDiskReadOnly(xcfg, 1));
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(chkImgRO2), cfgGetDiskReadOnly(xcfg, 2));
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(chkImgRO3), cfgGetDiskReadOnly(xcfg, 3));

	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(chkFastDiskDMA), cfgGetDiskFast(xcfg));

	/* ----- MEMORY PAGE ----- */
	ok = FALSE;
	for(i = 1; i <= 9; i++) {
		if(cfgGetChipSize(xcfg) == memSizes[i]) {
			gtk_entry_set_text(GTK_ENTRY(txtMemChipEntry), memTxt[i]);
			ok = TRUE;
		}
	}
	if(! ok) {
		/* should never happen */
		gtk_entry_set_text(GTK_ENTRY(txtMemChipEntry), memTxt[2]); /* default to 512KB */
		fellowAddLog("cfg2gui(): impossible value for cfgGetChipSize()\n");
	}

	sprintf(tmpstr, "%c MB", '0' + cfgGetFastSize(xcfg) / 0x100000);
	gtk_entry_set_text(GTK_ENTRY(txtMemFastEntry), tmpstr);

	ok = FALSE;
	for(i = 0; i <= 8; i++) {
		if(cfgGetBogoSize(xcfg) == memSizes[i]) {
			gtk_entry_set_text(GTK_ENTRY(txtMemBogoEntry), memTxt[i]);
			ok = TRUE;
		}
	}
	if(! ok) {
		/* should never happen */
		gtk_entry_set_text(GTK_ENTRY(txtMemBogoEntry), memTxt[0]); /* default to none */
		fellowAddLog("cfg2gui(): impossible value for cfgGetBogoSize()\n");
	}

	tmpstr = cfgGetKickImage(xcfg);
	gtk_entry_set_text(GTK_ENTRY(txtKickImage), tmpstr);
	tmpstr = cfgGetKey(xcfg);
	gtk_entry_set_text(GTK_ENTRY(txtKickKey), tmpstr);

	/* ----- SOUND PAGE ----- */
	switch(cfgGetSoundEmulation(xcfg)) {
	case SOUND_EMULATE:
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(radSndEmuSilent), TRUE);
		break;
	case SOUND_PLAY:
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(radSndEmuPlay), TRUE);
		break;
	case SOUND_NONE:
	default:
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(radSndEmuNo), TRUE);
	}

	/* filter */
	switch(cfgGetSoundFilter(xcfg)) {
	case SOUND_FILTER_ORIGINAL:
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(radFilterOrig), TRUE);
		break;
	case SOUND_FILTER_ALWAYS:
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(radFilterAlways), TRUE);
		break;
	case SOUND_FILTER_NEVER:
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(radFilterNever), TRUE);
	}

	/* rate */
	switch(cfgGetSoundRate(xcfg)) {
	case SOUND_44100:
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(radSamRate44100), TRUE);
		break;
	case SOUND_31300:
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(radSamRate31300), TRUE);
		break;
	case SOUND_22050:
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(radSamRate22050), TRUE);
		break;
	case SOUND_15650:
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(radSamRate15650), TRUE);
	}

	if(cfgGetSoundStereo(xcfg)) {
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(radChanStereo), TRUE);
	} else {
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(radChanMono), TRUE);
	}

	if(cfgGetSound16Bits(xcfg)) {
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(radBits16), TRUE);
	} else {
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(radBits8), TRUE);
	}

	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(chkSoundStreamWAV),
			cfgGetSoundWAVDump(xcfg));

	/* ----- SCREEN PAGE ----- */

	/* ----- FILESYSTEMS PAGE ----- */

	/* ----- HARDFILE PAGE ----- */

	/* ----- GAMEPORT PAGE ----- */
}
